#!
"""
Package for reading and analyzing Mide Instrumentation Data Exchange (MIDE)
files.
"""

__author__ = "David Randall Stokes"
__copyright__ = "Copyright (c) 2020 Mid\xe9 Technology"

__maintainer__ = "David Randall Stokes"
__email__ = "dstokes@mide.com"

__version__ = (3, 0, 0)
__status__ = "Development"

from .importer import importFile
